#class auto

#define BUF_SIZE	6144

#define SPI_SER_B
#define SPI_CLK_DIVISOR			5
#define SPI_CLOCK_MODE 1
#use SPI.LIB


#define MMC_SECTORSZ 512
#define MMC_SECTORL2 9

//data block transfer tokens
#define MMC_START_BLOCK_SREAD		0xFE
#define MMC_START_BLOCK_MREAD		0xFE
#define MMC_START_BLOCK_SWRITE	0xFE
#define MMC_START_BLOCK_MWRITE	0xFC
#define MMC_STOP_TRANS				0xFD

//responses
#define MMC_R1_OK      	0x00
#define MMC_R1_IDLING   0x01

//data responses
#define MMC_ACCEPTED		0x05
#define MMC_CRCERR		0x0B
#define MMC_WRERR			0x0D
// mask
#define MMC_DRMASK		0x1F
//busy signal
#define MMC_BUSY			0x00




// MMC SPI supported command set: name, CMD#
// (not necessarily supported on this implementation)
#define MMC_GO_IDLE_STATE				00
#define MMC_SEND_OP_COND				01
#define MMC_SEND_CSD						09
#define MMC_SEND_CID						10
#define MMC_STOP_TRANSMISSION			12
#define MMC_SEND_STATUS					13
#define MMC_SET_BLOCKLEN				16
#define MMC_READ_SINGLE_BLOCK			17
#define MMC_READ_MULTIPLE_BLOCK		18
#define MMC_WRITE_BLOCK					24
#define MMC_WRITE_MULTIPLE_BLOCK		25
#define MMC_PROGRAM_CSD					27
#define MMC_SET_WRITE_PROT				28
#define MMC_CLR_WRITE_PROT				29
#define MMC_SEND_WRITE_PROT			30
#define MMC_TAG_SECTOR_START			32
#define MMC_TAG_SECTOR_END				33
#define MMC_UNTAG_SECTOR				34
#define MMC_TAG_ERASE_GROUP_START	35
#define MMC_TAG_ERASE_GROUP_END		36
#define MMC_UNTAG_ERASE_GROUP			37
#define MMC_ERASE							38
#define MMC_READ_OCR						58
#define MMC_CRC_ON_OFF					59


// Timeout constants in bytes time
// NCR=max 8 bytes
#define MMC_TMOUT_NCR					8
// NAC=max (10*TAAC*f+100*NSAC)/8
#define MMC_TMOUT_NAC					1000
// wait after reset, technically not a timeout...
#define MMC_TMOUT_RST					10
// tired of getting busy signal (there is no spec about this)
#define MMC_TMOUT_BSY					1000


// Miscellaneous constants
#define DUMMY	0xFF
#define MMC_TIMEOUT     DUMMY
//CRC for CMD0
#define MMC_CMD0CRC 0x4A

#define MMC_TRANSFER_OK 0xAA

const static unsigned char DUMMYBUF[]={DUMMY,DUMMY};


/* Init related hardware
*/
void MMC_hardwareinit(void)
{
	BitWrPortI ( PBDDR, &PBDDRShadow, 1, 1 );	// CS = output
	BitWrPortI ( PEDDR, &PEDDRShadow, 1, 0 );	// PWR = output
	BitWrPortI ( PEFR, &PEFRShadow, 0, 0 );	// claim I/O pin
	BitWrPortI ( PBDR, &PBDRShadow, 0, 1 );	// CS=0
	BitWrPortI ( PEDR, &PEDRShadow, 0, 0 );	// no PWR
	BitWrPortI ( PCDR, &PEDRShadow, 0, 4 );	// PC4=0
};

/* Send a command to the MMC card, crc is 7-bit CRC right aligned
*/
nodebug void MMC_SndCmd (unsigned char cmd, unsigned long data, unsigned char crc)
{
unsigned char *a,buf[6];
int i;

	a=(unsigned char *) &data;		// LSB first, pay attention to endianness if porting
	buf[0]=cmd|0x40;					// add transmitter bit for host command, start bit is 0 because of cmd number
	i=4;
	do {
		buf[i]=(*(a++));				// MSB first, pay attention to endianness if porting
   } while(--i);
	buf[5]=((crc<<1)|0x01);			// add end bit
   SPIWrite(buf,6);					// Send
}

/* get a response from the MMC card, and return it
Timeout must be >0, on timeout, last response is DUMMY,
so on timeout, returned value is effectively DUMMY, which is MMC_TIMEOUT
*/
nodebug char MMC_GetResponse(int timeout)
{
unsigned char response;

	while(timeout--){
   	SPIRead(&response,1);
		if(response!=DUMMY)
			break;
	}
	return(response);
}


/* Busy waits until card reports init OK, error, or timeout
Returns MMC_R1_OK on success, MMC_R1_IDLING on timeout, error R1 on error
*/
unsigned char MMC_init (unsigned int timeout)
{
unsigned char response;
int t;

	BitWrPortI ( PBDR, &PBDRShadow, 1, 1 );	// disable CS
	t=MMC_TMOUT_RST;						// wait some time after hardware reset
   while(t--)
		SPIWrite(DUMMYBUF,1);
	BitWrPortI ( PBDR, &PBDRShadow, 0, 1 );	// enable CS

	MMC_SndCmd(MMC_GO_IDLE_STATE,0L,MMC_CMD0CRC);	// reset card
/* after getting response to MMC_GO_IDLE_STATE, repeatedly poll
card until it responds MMC_R1_OK or we get tired, whatever comes first */
	while((response=MMC_GetResponse(MMC_TMOUT_NAC))==MMC_R1_IDLING){
      if(!timeout--)
      	break;
		BitWrPortI ( PBDR, &PBDRShadow, 1, 1 );	// disable CS
		SPIWrite(DUMMYBUF,1);
		BitWrPortI ( PBDR, &PBDRShadow, 0, 1 );	// enable CS
		MMC_SndCmd(MMC_SEND_OP_COND,0L,DUMMY);			// start/poll initialization process
   }
	BitWrPortI ( PBDR, &PBDRShadow, 1, 1 );	// disable CS
	SPIWrite(DUMMYBUF,1);			// keep clock going for some time
   printf("Response: 0x%02X\n",response);
	return (response);
}

/* Returns after busy signal is removed, or timeout
Returns read value if != MMC_BUSY
*/
nodebug unsigned char MMC_bsywait(int timeout)
{
unsigned char response;

	while(timeout--){
		SPIRead(&response,1);
		if(response!=MMC_BUSY)
   	 	break;
  }
  return(response);
}

/* Read a block of data, leave it at buf
Returns MMC_TRANSFER_OK if OK, errored R1 or data token on error
Not very friendly on error, feel free to modify if you really need to know
what is wrong
*/
nodebug unsigned char MMC_rdblk(unsigned char *buf,unsigned int len,)
{
unsigned char response,CRCbuf[2];

	if ((response=MMC_GetResponse(MMC_TMOUT_NCR))==MMC_R1_OK) {
		if ((response=MMC_GetResponse(MMC_TMOUT_NAC))==MMC_START_BLOCK_SREAD) {
			SPIRead(buf,len);       // get block
			SPIRead(CRCbuf,2);		// get CRC (ignored)
			response=MMC_TRANSFER_OK;
		}
	}
   return(response);
}


/* Read a sector, leave data at buf
uses MMC_rdblk, so see it
*/
nodebug unsigned char MMC_RdSector(unsigned long sector,unsigned char *buf)
{
unsigned char response;
int i;
unsigned long address;

	address=sector << MMC_SECTORL2;
	BitWrPortI ( PBDR, &PBDRShadow, 0, 1 );	// enable CS
	MMC_SndCmd(MMC_READ_SINGLE_BLOCK,address,DUMMY);
   response=MMC_rdblk(buf,MMC_SECTORSZ);
	BitWrPortI ( PBDR, &PBDRShadow, 1, 1 );	// disable CS
	SPIWrite(DUMMYBUF,1);
	return(response);
}




unsigned char buffer[MMC_SECTORSZ];
unsigned long fat_offset,root_dir_offset,data_area_offset;
unsigned int sectors_cluster;


/* REDUCED FAT FAST routines */

#asm
FIND::
		LD IY,buffer
FLOP: BIT 3,(IY+11)
      JR NZ,FLOP1
      BIT 2,(IY+11)
      JR NZ,FLOP1
      LD A,(IY)
      AND A
      JR Z,NONE
      CP (HL)
      JR NZ,FLOP1
      CALL GO1
      JR NC,FLOP1
      PUSH HL
      PUSH IY
      LD L,C
      LD H,B
      LD BC,7
      ADD IY,BC
      LD B,3
      CALL GO2
      POP IY
      POP HL
      JR NC,FLOP1
      LD HL,IY
;		AND A
;      LD DE,buffer
;      SBC HL,DE
      RET
FLOP1: LD DE,32
		ADD IY,DE
      EX DE,HL
      LD HL,IY
      LD BC,buffer+512
      AND A
      SBC HL,BC
      EX DE,HL
      JR C,FLOP
NONE: LD HL,-1
      RET
;
GO1:  LD B,7
GO2:  PUSH IY
      PUSH HL
GLOP: INC HL
      INC IY
      LD A,(IY)
      CP (HL)
      JR NZ,DIF1
      DJNZ GLOP
      INC HL
IGU:  LD C,L
      LD B,H
      POP HL
      POP IY
      SCF
      RET
DIF1: CP 0x20
      JR NZ,DIF
      LD A,'.'
      CP (IY-1)
      JR Z,DIF
      CP (HL)
      JR Z,IGU
DIF:  POP HL
      POP IY
      AND A
      RET
#endasm


int MMCFAT_init()
{
	MMC_hardwareinit();
	BitWrPortI ( PEDR, &PEDRShadow, 1, 0 );	// PWR
	SPIinit();
	if(MMC_init(5000)!=MMC_R1_OK){
      return(-1);
	}
   return(1);
}

int MMCFAT_mount(unsigned int part)
{
long addr;

	if(part>3)
   	return(-1);
	if(MMC_RdSector(0L, buffer)==MMC_TRANSFER_OK){
        	printf("File system descriptor for first partition: ");
   		switch(buffer[0x1be+0x10*part+4]){
   		case 4:
         	printf("DOS 16-bit FAT <32MB\n");
            break;
   		case 6:
         	printf("DOS 16-bit FAT >32MB\n");
            break;
			default:
         	printf("Unsupported\n");
            return(-1);
         }
          addr=(*((long *)&buffer[0x1be+0x10*part+8]));
			 printf("Address of first sector in first partition: %08lX\n\n",addr);
		 	 if(MMC_RdSector(addr, buffer)==MMC_TRANSFER_OK){
          	buffer[0xb]=0;
            printf("OEM name: %s\n",&buffer[3]);
          	buffer[0x3e]=0;
            printf("Volume label, file system type: %s\n",&buffer[0x2b]);
            sectors_cluster=buffer[0x0D];
            fat_offset=addr+*((int *)&buffer[0x0e]);
            root_dir_offset=fat_offset+(buffer[0x10]*(*((int *)&buffer[0x16])));
            data_area_offset=root_dir_offset+((long)*((int *)&buffer[0x11])*32)/512;
            printf("Sectors per cluster: %d\n",sectors_cluster);
            printf("Root directory entries: %d\n",*((int *)&buffer[0x11]));
            printf("FAT first sector: %ld\n",fat_offset);
            printf("Root directory first sector: %ld\n",root_dir_offset);
            printf("Data area first sector: %ld\n",data_area_offset);
            return(1);
          }
	}
   return(-1);
}

static unsigned long offset_in_file,file_lenght;
static unsigned int sector_in_cluster;
static unsigned int first_file_cluster,cur_file_cluster,cur_file_offset;

nodebug void MMCFAT_rewind(void)
{
      cur_file_cluster=first_file_cluster;
      cur_file_offset=0;
      sector_in_cluster=0;
      offset_in_file=0;
}

int MMCFAT_open(char *name)
{
int entry;
long sector;

	sector=root_dir_offset;
   do {
		if(MMC_RdSector(sector++, buffer)==MMC_TRANSFER_OK){
			if((entry=FIND(name))!=-1){
		   	if(!(*(unsigned char *)(entry+11)&0x10)){	// not a directory entry
               file_lenght=*((long *)(entry+28));
               printf("%ld bytes\n",file_lenght);
               first_file_cluster=*(int *)(entry+26);
               MMCFAT_rewind();
			   	return(first_file_cluster);
            }
         }
		}
   } while(sector < data_area_offset);
	return(-1);
}

#define MMCFAT_cluster2sector(i)	((unsigned long)sectors_cluster*(i-2)+data_area_offset)

nodebug int MMCFAT_nextcluster(int cluster)
{
long FATsector;
int FATentry;

	FATsector=fat_offset+(((long)cluster)>>8);	// 2*cluster/512, <<1 >>9
   FATentry=(cluster<<1)&0x01FF;						// (2*cluster)%512
	if(MMC_RdSector(FATsector, buffer)!=MMC_TRANSFER_OK)
		return(-1);
   return(buffer[FATentry]+(buffer[FATentry+1]<<8));
}

nodebug int MMCFAT_seek(long pos)
{
unsigned int clusters,curcluster,newcluster,sectors,bytes;
unsigned char *bptr;

	if(pos<0)
		MMCFAT_rewind();
	if(pos==-1L)
   	pos=0;
	sectors=(int)(pos >> 9);				// /bytes per sector
   bytes=(int)(pos&0x1FF);					// %bytes per sector
   clusters=sectors/sectors_cluster;
   sectors%=sectors_cluster;
   bytes+=cur_file_offset;
   sectors+=sector_in_cluster;
	if(bytes>=512){
      bytes-=512;
      sectors++;
   }
   if(sectors>=sectors_cluster){
   	sectors-=sectors_cluster;
      clusters++;
   }
   newcluster=curcluster=cur_file_cluster;
	while(clusters--){
	   if((newcluster=MMCFAT_nextcluster(curcluster))==-1){
      	if(pos<0L){
         		offset_in_file=file_lenght;
            	break;
         }
         else
         	return(-1);
      }
      curcluster=newcluster;
   }
	if(MMC_RdSector(MMCFAT_cluster2sector(curcluster)+sectors, buffer)!=MMC_TRANSFER_OK)
           	return(-1);
	cur_file_cluster=curcluster;
   sector_in_cluster=sectors;
   cur_file_offset=bytes;
	if(offset_in_file!=file_lenght)
   	offset_in_file+=pos;
   return(1);
}

nodebug int MMCFAT_read(unsigned char *buf,int bytes)
{
int count,helpme;
unsigned char *bptr;

	if(offset_in_file==0L){
		if(MMCFAT_seek(0)==-1)
  	     	return(-1);
   }
	count=bytes;
   bptr=buffer+cur_file_offset;
	while(bytes&&(offset_in_file<file_lenght)){
		if(cur_file_offset>=512){
	      cur_file_offset=0;
         bptr=buffer;
      	if(++sector_in_cluster==sectors_cluster){
         	sector_in_cluster=0;
            if((cur_file_cluster=MMCFAT_nextcluster(cur_file_cluster))==-1)
            	break;
         }
			if((helpme=MMC_RdSector(MMCFAT_cluster2sector(cur_file_cluster)+sector_in_cluster, buffer))!=MMC_TRANSFER_OK)
           	break;
   	}
      *(buf++)=*(bptr++);
      cur_file_offset++;
      offset_in_file++;
      bytes--;
   }
	return(count-bytes);
}

nodebug int MMCFAT_fastread(unsigned char *buf,int bytes)
{
int count,bbytes;
long bbbytes;

	if(offset_in_file==0L){
		if(MMCFAT_seek(0)==-1)
  	     	return(-1);
   }
	count=bytes;
   bbbytes=file_lenght-offset_in_file;
   if(bbbytes<0)
	   return(0);
	while(bytes&&bbbytes){
		if(cur_file_offset>=512){
	      cur_file_offset=0;
      	if(++sector_in_cluster==sectors_cluster){
         	sector_in_cluster=0;
            if((cur_file_cluster=MMCFAT_nextcluster(cur_file_cluster))==-1)
            	break;
         }
			if(MMC_RdSector(MMCFAT_cluster2sector(cur_file_cluster)+sector_in_cluster, buffer)!=MMC_TRANSFER_OK)
           	break;
   	}
      bbytes=(bytes<512)? bytes:512;
      bbbytes=file_lenght-offset_in_file;
		if(bbbytes<512)
      	bbytes=(int)bbbytes;
      memcpy(buf,buffer,bbytes);
      buf+=bbytes;
      cur_file_offset+=bbytes;
      offset_in_file+=bbytes;
      bytes-=bbytes;
   }
	return(count-bytes);
}


int printsector(long sector)
{
unsigned char rc;
int i;

	if((rc=MMC_RdSector(sector, buffer))==MMC_TRANSFER_OK){
	   for(i=0;i<MMC_SECTORSZ;i++){
      	if((i&0xFF0)==i)
   			printf("\n%04X\t",i);
   		printf("%02X ",buffer[i]);
	   }
   }
   else printf("result: %02X\n",rc);
   return(rc);
}

unsigned char buf1[BUF_SIZE],buf2[BUF_SIZE],*ptr;
unsigned int bytes,bytes1,bytes2;
unsigned char which,load,run;

#asm root
PWM_handler::
	push af
   ld a,(run)
   and a
   jr nz,ok
   ld a, 0x80
   ioi ld (PWM2R),a
   jr chau
ok:
   push hl
	ld hl,(ptr)
	ld a,(hl)
   ioi ld (PWM2R),a
   inc hl
   ld (ptr),hl
   ld hl,(bytes)
   dec hl
   ld (bytes),hl
   ld a,h
   or l
   jr nz,done
   ld a,1
   ld (load),a
   ld a,(which)
   xor 0xFF
   ld (which),a
   jr nz,use2
	ld hl,buf1
   ld (ptr),hl
   ld hl,(bytes1)
   jr use
use2:
	ld hl,buf2
   ld (ptr),hl
   ld hl,(bytes2)
use:
   ld (bytes),hl
   ld a,h
   or l
   jr nz,done
   ld (run),a
done:
	pop hl
chau:
	pop af
	ipres
   ret
#endasm

int main( void )
{
unsigned char rc;
int i,k;
char filename[13];


	if(MMCFAT_init()==-1)
   	goto oops;
	if(MMCFAT_mount(0)==-1)
   	goto oops;

   run=0;
	pwm_init(28800L);
   SetVectIntern(0x17,PWM_handler);
   pwm_set(2,513,PWM_SPREAD | PWM_INTNORMAL | PWM_USEPORTC | PWM_INTPRI3);

	for(k=1;k<4;k++){
      sprintf(filename,"FORD%d.PCM",k);
      puts(filename);
	   if(MMCFAT_open(filename)!=-1) {
	      which=load=1;
	   	do {
				if(load){
   	         load=0;
	           	if(which){
				   	i=MMCFAT_fastread(buf1,BUF_SIZE);
//                  puts("1");
                  bytes1=i;
					}
//               else puts("-");
	           	if((!which)||(!run)){
				   	i=MMCFAT_fastread(buf2,BUF_SIZE);
//                  puts("2");
                  bytes2=i;
					}
//               else puts("-");
               if(!run){
				      which=0;
                  bytes=bytes1;
					   ptr=buf1;
               	run=1;
               }
				}
	      } while (i>0);
        	if(which)
         	bytes1=0;
         else
         	bytes2=0;
      }
		while(run);
   }

oops:	BitWrPortI ( PEDR, &PEDRShadow, 0, 0 );	// no PWR
}


